﻿using System;
using System.Text.RegularExpressions;
using Excel = Microsoft.Office.Interop.Excel;

namespace DmisLib.Calc
{
    public class ExcelDocument : IBibReport
    {
        private Excel.Application xlApp;
        private Excel.Workbook xlWorkBook;
        private Excel.Worksheet xlWorkSheet;

        /// <summary>
        /// create the excel document object without opening an excel document.
        /// </summary>
        public ExcelDocument()
        {
            xlApp = new Excel.Application();
            xlApp.Visible = false;
        }

        /// <summary>
        /// Create the excel document object and open an excel document.
        /// </summary>
        /// <param name="ExcelFile"></param>
        public ExcelDocument(String ExcelFile)
            : this()
        {
            Open(ExcelFile);
        }

        /// <summary>
        /// Open the Excel Template for the report.
        /// </summary>
        /// <param name="excelFileName"></param>
        public void Open(string excelFileName)
        {
            xlWorkBook = xlApp.Workbooks.Open(excelFileName);
            SelectSheet(1);
        }

        /// <summary>
        /// Select a worksheet, 1 is the first sheet.
        /// </summary>
        /// <param name="sheetNumber"></param>
        public void SelectSheet(int sheetNumber)
        {
            if (xlWorkBook != null)
            {
                xlWorkSheet = xlWorkBook.Worksheets.Item[sheetNumber] as Excel.Worksheet;
            }
        }

        /// <summary>
        /// Close the excel document.
        /// </summary>
        public void Close()
        {
            xlWorkBook.Close(SaveChanges: false);
            xlApp.Quit();
            
            releaseObject(xlWorkSheet);
            releaseObject(xlWorkBook);
            releaseObject(xlApp);

            xlWorkSheet = null;
            xlWorkBook = null;
            xlApp = null;

        }

        /// <summary>
        /// Save the Excel document as a PDF file.
        /// </summary>
        /// <param name="fileName"></param>
        public void SaveAsPdfDocument(string fileName)
        {
            xlWorkBook.ExportAsFixedFormat(Excel.XlFixedFormatType.xlTypePDF, fileName + ".pdf");
        }

        /// <summary>
        /// Save the Excel document as an XPS file.
        /// </summary>
        /// <param name="fileName"></param>
        public void saveAsXpsDocument(string fileName)
        {
            xlWorkBook.ExportAsFixedFormat(Excel.XlFixedFormatType.xlTypeXPS, fileName + ".xps");
        }

        /// <summary>
        /// Save the Excel document.
        /// </summary>
        /// <param name="fileName"></param>
        public void SaveAsExcelDocument(string fileName)
        {
            xlWorkBook.SaveAs(fileName + ".xls");
        }

        /// <summary>
        /// Write a string value into a cell.
        /// </summary>
        /// <param name="cellAddress"></param>
        /// <param name="content"></param>
        public void LocalCell(String cellAddress, String content)
        {
            if (content == "") return;
            if (content.StartsWith("="))
            {
                // this is probably a formula
                xlWorkSheet.Range[cellAddress].Value = content;
            }
            else
            {
                // this is a text
                xlWorkSheet.Range[cellAddress].Value = content;
            }
        }
        
        /// <summary>
        /// Write a double value into a cell.
        /// </summary>
        /// <param name="cellAddress"></param>
        /// <param name="content"></param>
        public void LocalCell(String cellAddress, Double content)
        {
            // this is a double value
            xlWorkSheet.Range[cellAddress].Value = content;
        }

        /// <summary>
        /// Get the String contents of a cell.
        /// </summary>
        /// <param name="cellAddress"></param>
        /// <param name="type"></param>
        /// <param name="content"></param>
        public String LocalCell(String cellAddress, CELL_CONTENT type)
        {
            Excel.Range sheetCell = xlWorkSheet.Range[cellAddress];

            String content = sheetCell.Value2.ToString();

            switch (type)
            {
                case CELL_CONTENT.STRING:
                    content = sheetCell.Value2.ToString();
                    break;
                case CELL_CONTENT.NUMBER:
                if (sheetCell.Value2 != null)
                {
                    content = sheetCell.Value2.ToString();
                }
                    break;
                case CELL_CONTENT.FORMULA:
                    content = sheetCell.Formula.ToString();
                    break;
                default:
                    break;
            }

            return content;

        }

        private void releaseObject(Object obj)
        {
            try
            {
                System.Runtime.InteropServices.Marshal.FinalReleaseComObject(obj);
                obj = null;
            }
            catch (Exception ex)
            {
                String message = ex.Message;
                obj = null;
            }
            finally
            {
                GC.Collect();
            }
        }
    }
}
