﻿using System;
using System.Linq;
using System.IO;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Microsoft.Win32;
using DmisLib.Geometry;
using PCDLRN;


namespace DmisLib.Session
{
    using pcdOBT = PCDLRN.OBTYPE;
    using pcdEFT = PCDLRN.ENUM_FIELD_TYPES;

    public enum CONTROL_OPTION
    {
        STOP,
        CONTINUE,
        STEPNEXT,
        STEPINTO,
        SKIP,
        ERASEHIT,
        VISABLE,
        NOT_VISABLE
    }

    public enum UNIT_OPTION
    {
        IN,
        MM
    }

    /// <summary>
    /// PcdmisHelper is an interface between your code and PC-Dmis.
    /// Using this class library it is not necessary for your C# code
    /// to include a reference to PCDLRN.
    /// </summary>
    public class PcdmisHelper
    {
        // Main object which represents the PCDLRN session.
        private PCDLRN.Application pcdSession;
        // Collection of PCDMIS part programs.
        private PCDLRN.PartPrograms pcdPartPrograms;
        // One part program.
        private PCDLRN.PartProgram pcdPartProgram;
        // Collection of PCDMIS commands within a part program.
        private PCDLRN.Commands pcdCommands;
        // A single PCDMIS command.
        private PCDLRN.Command pcdCommand;
        // PCDMIS Class identifier
        private string PCDMIS_CLSID;

        // Full pathname to PCDMIS folder.
        private string pcdDefaultPcdmisPath;
        // Full path to the probe folder
        private string pcdProbeFolderPath;

        // Version of PCDMIS to connect
        private string PcdmisVersion;

        // Measurement unit of the programs
        private UNITTYPE MeasurementUnit;

        // Name of CMM.
        private string pcdDefaultMachine = "CMM1";

        // PCDMIS Control window
        private ExecutionWindow pcdControl;

        /* Dimension Factory unpacks PCDMIS dimension command into a 
         * DmisLib.Session.DimensionData Object which is returned to the application using
         * PcdmisHelper. This is done so that the application using PcdmisHelper
         * does not have to add a reference to the PCDMIS PCDLRN. */
        private DimensionFactory dimFactory;

        /* The following event delegates are used to send event signals back
         * to the user application while PCDMIS is measuring a part.*/

        // Triggers when a connection is completed to the PCDLRN object.
        public delegate void ConnectionEventHandler(Boolean results);
        public event ConnectionEventHandler ConnectedEvent;

        // Triggers when executing program has terminated.
        public delegate void TerminationEventHandler(int results);
        public event TerminationEventHandler LocalProgramTerminationEvent;

        // Triggers when a Dimension object has been processed.
        public delegate void DimensionEventHandler(DimensionData data);
        public event DimensionEventHandler DimensionEvent;

        // Triggers when a 'Show' comment has been processed.
        public delegate void ShowEventHandler(string label);
        public event ShowEventHandler ShowEvent;

        public PcdmisHelper(string applicationName, string Version, UNIT_OPTION unit) : this(Version, unit)
        {
            ApplicationFolderName = applicationName;
        }

        // Constructor
        /// <summary>
        /// Contructor to create PcdmisHelper object passing a string to indicate which
        /// version of PCDMIS to connect. Current valid version strings are:
        /// 2011 MR1, 2012, 2012 MR1, 2013, 2015, 2015.1
        /// </summary>
        /// <param name="version"></param>
        /// <param name="unit"></param>
        /// </summary>
        public PcdmisHelper(string version, UNIT_OPTION unit)
        {
            // Init variables with default values.
            PcdmisVersion = version;
            InitHelper();

            MeasurementUnit = UNITTYPE.MM; // default units is millimeters
            if (unit == UNIT_OPTION.IN) MeasurementUnit = UNITTYPE.INCH;

            /* DimensionFactory is used by PcdmisHelper to convert the PCDMIS dimension 
            * to a DmisLib DimensionData data type.
            * As of PCDMIS 2010 or 2011, PCDMIS dimensions have a new structure to accomodate
            * feature controls.  At this time DimensionFactory is not able to parse out
            * the data from this type of structure.  It is necessary for the PCDMIS programer
            * when preparing the PCDMIS program to select the option to use old style legency dimensions */
            dimFactory = new DimensionFactory();

            /* Pcdmis version selection - Input string must be one of the following
             * "2011 MR1" = Internal version 6.1
             * "2012" = Internal version 7.0
             * "2012 MR1" = Internal version 7.1
             * "2013" = Internal version 8.0
             * "2015" = (64-bit) Internal version 10.0
             * "2015.1 = (64-bit) Internal version 10.1 
             * Default is the last version that was executed */

            /* To properly connect with PC-Dmis you need the internal version number.
             * To access PCDMIS register settings such as ProbeFolderPath etc. you need the external version.
             * This switch maps the external version with the internal version to assure that the correct version
             * of PCDLRN and that the correct local registry settings for the selected version are used.*/

            version = "V" + version;
            switch (version)
            {
                case "V2011 MR1":
                    PCDMIS_CLSID = getCLSID_RegisterValue("6.1");
                    break;
                case "V2012":
                    PCDMIS_CLSID = getCLSID_RegisterValue("7.0");
                    break;
                case "V2012 MR1":
                    PCDMIS_CLSID = getCLSID_RegisterValue("7.1");
                    break;
                case "V2013":
                    PCDMIS_CLSID = getCLSID_RegisterValue("8.0");
                    break;
                case "V2013 MR1":
                    PCDMIS_CLSID = getCLSID_RegisterValue("8.1");
                    break;
                case "V2015":
                    PCDMIS_CLSID = getCLSID_RegisterValue("10.0");
                    break;
                case "V2015.1":
                    PCDMIS_CLSID = getCLSID_RegisterValue("10.1");
                    break;
                default:
                    PCDMIS_CLSID = getCLSID_RegisterValue("");
                    break;
            }
        }

        private string getCLSID_RegisterValue(string Parameter)
        {
            // Set a register key for version
            string regValue = null;
            string PcdmisClassID;
            if (PcdmisVersion == "")
            {
                PcdmisClassID = @"PCDLRN.Application\CLSID";
            }
            else
            {
                PcdmisClassID = @"PCDLRN.Application." + Parameter + @"\CLSID";
            }

            RegistryKey reg1 = Registry.ClassesRoot.OpenSubKey(PcdmisClassID);
            if (reg1 != null)
            {
                regValue = reg1.GetValue("") as string;
                if (regValue != null)
                {
                    reg1.Close();
                    return regValue;
                }
            }
            reg1.Close();

            return "";
        }

        private void InitHelper()
        {
            SubroutineFolderName = "Subroutines";
            pcdDefaultPcdmisPath = getRegisterValue("DefaultPartProgram");
            pcdProbeFolderPath = getRegisterValue("LoadProbeUserSearchDirectory");
        }

        private string getRegisterValue(string Parameter)
        {
            // Set a register key for version
            string regValue = null;
            string baseKey1 = @"SOFTWARE\WAI\PC-DMIS\" + PcdmisVersion + @"\Option";
            string baseKey2 = @"SOFTWARE\WAI\PC-DMIS\" + PcdmisVersion + @"\USER_Option";

            RegistryKey reg1 = Registry.CurrentUser.OpenSubKey(baseKey1);
            if (reg1 != null)
            {
                regValue = reg1.GetValue(Parameter) as string;
                if (regValue != null)
                {
                    reg1.Close();
                    return regValue;
                }
                reg1.Close();
            }

            RegistryKey reg2 = Registry.LocalMachine.OpenSubKey(baseKey2);
            if (reg2 != null)
            {
                regValue = reg2.GetValue(Parameter) as string;
                if (regValue != null)
                {
                    reg2.Close();
                    return regValue;
                }
                reg2.Close();
            }

            return "ParameterNotFound";
        }

        public string LastProbeFile
        {
            get { return getRegisterValue("LastProbeFile") as string; }
        }

        // Create on a separate thread an active connection to PCDMIS/PCDLRN Object.
        public void connectPCDLRN()
        {
            if (!IsSessionActive)
            {
                Func<Boolean> method = connect;
                method.BeginInvoke(connectionDone, method);
            }
            else
            {
                // if user calls this method and the session is already connected then
                // the user copy of PcdmisHelper will be reset to known state per sessionSetup()
                // To avoid the reset of this instance of PcdmisHelper check for connectivity
                // using SessionIsActive property.
                initSession();

                // Fire the connected event if there are any interested listeners
                if (ConnectedEvent != null) { ConnectedEvent(IsSessionActive); }
            }
        }

        /* execute this method to make connection to the PCDLRN. */
        private Boolean connect()
        {
            Type comType;
            Object comObj;

            try
            {
                pcdSession = null;
                // Make sure the PCDLRN process is not already running.
                Process[] PCDLRN_List = Process.GetProcessesByName("PCDLRN");
                foreach (Process exe in PCDLRN_List)
                {
                    exe.Kill();
                }
                comType = Type.GetTypeFromCLSID(Guid.Parse(PCDMIS_CLSID), true);
                comObj = Activator.CreateInstance(comType);
                pcdSession = comObj as PCDLRN.Application;
                return IsSessionActive;
            }
            catch (COMException e)
            {
                string msg = e.Message;
                return IsSessionActive;
            }
        }

        // After the connection is made with Pcdmis the calling program is notified.
        private void connectionDone(IAsyncResult cookie)
        {
            var target = (Func<Boolean>)cookie.AsyncState;
            Boolean result = target.EndInvoke(cookie);

            if (IsSessionActive)
            {
                // Setup the session to a known state
                initSession();

                // if there are any interested listeners... let them know that we are connected
                if (ConnectedEvent != null) ConnectedEvent(result);
            }
        }

        /* This is called after a connection is made. Its purpose is to place this 
         * instance of the PcdmisHelper in a known state with no open programs. */
        private void initSession()
        {
            // Put the session in a known state: Available with no part programs open or active.
            SessionIsAvailable = true;
            pcdPartPrograms = pcdSession.PartPrograms;
            if (pcdPartPrograms.Count > 0)
            {
                pcdPartProgram = pcdSession.ActivePartProgram;
                pcdPartProgram.Quit();
                pcdPartPrograms.CloseAll();
            }
            pcdPartProgram = pcdSession.ActivePartProgram;
            if (pcdPartProgram != null) pcdPartProgram.Quit();            
        }

        //PCDMIS Part Program event handler
        void pcdPartProgram_OnObjectAboutToExecute(PCDLRN.Command cmd)
        {
            string label;
            if (IsShowCommand(cmd,out label))
            {
                if (ShowEvent != null) ShowEvent(label);
            }
        }

        private Boolean IsShowCommand(PCDLRN.Command cmd, out string label)
        {
            /* This code is part of an experiment. In the PC-Dmis program the programmer places a
             * comment with the first word "Show" and the second word a label. The label is passed
             * back to the DotNet program which uses the label to identfy and display the info 
             * relevent to the operator at that point in the program. The purpose of this is primarily to leverage
             * the DotNet ability to handle international language*/
            label = "";
            if (cmd.Marked)
            {
                if (cmd.IsComment)
                {
                    PCDLRN.Comment CmdComment = cmd.CommentCommand;
                    if (CmdComment.Comment.Contains("Show"))
                    {
                        Char[] delimiters = { ' ' };
                        string comment = CmdComment.Comment;
                        string[] words = comment.Split(delimiters, StringSplitOptions.RemoveEmptyEntries);
                        if (words.Length > 1)
                        {
                            label = words[1];
                            return true;
                        }
                    }
                }
            }

            return false;
        }

        //PCDMIS Part Program Event Handler
        void pcdPartProgram_OnObjectExecuted(PCDLRN.Command cmd)
        {
            if (cmd.Marked)
            {
                if (cmd.IsDimension)
                {
                    DimensionData dimData;
                    if (dimFactory.processDimension(cmd, out dimData))
                    {
                        if (DimensionEvent != null) DimensionEvent(dimData);
                    }
                    return;
                }

                if (cmd.IsComment)
                {
                    string label;
                    if (IsShowCommand(cmd, out label))
                    {
                        if (ShowEvent != null) ShowEvent(label);
                        return;
                    }
                }
            }
        }

        //PCDMIS Part Program Event Handler
        void pcdPartProgram_OnEndExecution(int TerminationType)
        {
            SessionIsAvailable = true;
            if (LocalProgramTerminationEvent != null) LocalProgramTerminationEvent(TerminationType);
        }


        /// <summary>
        /// Close all previously opened programs.
        /// </summary>
        public void CloseAll()
        {
            if (SessionIsAvailable)
            {
                try
                {
                    pcdPartPrograms.CloseAll();
                }
                catch (Exception ex)
                {
                    throw new UnknownPcdlrnErrorException(ex.Message);
                }
            }
        }

        /// <summary>
        /// Execute a PCDMIS program in different thread from calling program.
        /// </summary>
        /// <param name="program">Name of program without path or extension</param>
        /// <returns></returns>
        public bool executeAsyncProgram(string program)
        {
            if (SessionIsAvailable)
            {
                if (validateProgram(program))
                {
                    // program does exist in application folder.
                    SessionIsAvailable = false;
                    try
                    {
                        pcdPartPrograms.CloseAll();
                        pcdPartProgram = pcdPartPrograms.Open(AppProgFullName(program),
                            pcdDefaultMachine);

                        // assign event handlers to Pcdmis program events
                        pcdPartProgram.OnObjectAboutToExecute += new PCDLRN.IPartProgramEvents_OnObjectAboutToExecuteEventHandler(pcdPartProgram_OnObjectAboutToExecute);
                        pcdPartProgram.OnObjectExecuted += new PCDLRN.IPartProgramEvents_OnObjectExecutedEventHandler(pcdPartProgram_OnObjectExecuted);
                        pcdPartProgram.OnEndExecution += new PCDLRN.IPartProgramEvents_OnEndExecutionEventHandler(pcdPartProgram_OnEndExecution);

                        pcdControl = pcdPartProgram.GetExecutionWindow(0);
                        return pcdPartProgram.AsyncExecute();
                    }
                    catch (Exception ex)
                    {
                        throw new InvalidProgramException(ex.Message);
                    }
                }
                else
                {
                    throw new InvalidProgramException();
                }
            }

            return false;
        }

        /// <summary>
        /// Open and prepare a subprogram to receive assignments, tolerances and points
        /// to be written into it.
        /// </summary>
        /// <param name="subName">Name of subprogram without path or extension.</param>
        public Boolean OpenNewSubprogram(string subName)
        {
            Boolean retStatus;

            if (SessionIsAvailable)
            {
                SessionIsAvailable = false;
                pcdPartPrograms.CloseAll();
                // Make sure the original is deleted before initializing a new program.
                purgeProgram(subName);

                // construct Subroutine folder if there is supposed to be one
                if (!Directory.Exists(AppSubrDirName()))
                {
                    Directory.CreateDirectory(AppSubrDirName());
                }

                initNewPartProgram(AppProgSubrName(subName));

                // Add the subroutine command.
                pcdCommand = pcdCommands.Add(pcdOBT.START_SUBROUTINE, true);
                pcdCommand.Marked = true;
                retStatus = pcdCommand.PutText(subName, pcdEFT.SUB_NAME, 0);
                return retStatus;
            }

            return false;
        }

        /// <summary>
        /// Close a previously opened subprogram.
        /// </summary>
        public void CloseSubProgram()
        {
            pcdCommand = pcdCommands.Add(pcdOBT.END_SUBROUTINE, true);
            pcdCommand.Marked = true;

            pcdPartProgram.Save();
            pcdPartPrograms.CloseAll();
            SessionIsAvailable = true;
        }

        /// <summary>
        /// Send a signal to active program
        /// ToDo: This hasn't been tested
        /// </summary>
        public void ProgramControl(CONTROL_OPTION option)
        {
            if (pcdControl != null)
            {
                if (pcdControl.Available)
                {
                    switch (option)
                    {
                        case CONTROL_OPTION.STOP:
                            pcdControl.stop();
                            break;
                        case CONTROL_OPTION.CONTINUE:
                            pcdControl.Continue();
                            break;
                        case CONTROL_OPTION.STEPNEXT:
                            pcdControl.StepNext();
                            break;
                        case CONTROL_OPTION.STEPINTO:
                            pcdControl.StepInto();
                            break;
                        case CONTROL_OPTION.SKIP:
                            pcdControl.Skip();
                            break;
                        case CONTROL_OPTION.ERASEHIT:
                            pcdControl.EraseHit();
                            break;
                        case CONTROL_OPTION.VISABLE:
                            pcdControl.Visible = true;
                            break;
                        case CONTROL_OPTION.NOT_VISABLE:
                            pcdControl.Visible = false;
                            break;
                        default:
                            break;
                    }
                }
            }
        }

   // Probe methods follow:
        /// <summary>
        /// Set the default probe name to be used by PCDMIS programs.
        /// </summary>
        string pcdDefaultProbeName;
        public string DefaultProbeName
        {
            get { return pcdDefaultProbeName; }
            set { pcdDefaultProbeName = value; }
        }

        /// <summary>
        /// Builds and returns a list of probe names found in the pcd_Probes folder
        /// without the .PRB extension.
        /// </summary>
        /// <returns>String array containing probe names.</returns>
        public string[] ProbeList()
        {
            if (Directory.Exists(pcdProbeFolderPath))
            {
                try
                {
                    string[] p1List = Directory.GetFiles(pcdProbeFolderPath, "*.PRB");
                    string[] p2List = new string[p1List.Count()];

                    int i = 0;
                    foreach (string pname in p1List)
                    {
                        // Strip off the .prb extension from each file name
                        // before returning the list.
                        FileInfo fi = new FileInfo(pname);
                        int loc = fi.Name.IndexOf(".PRB", 0, fi.Name.Length,
                            StringComparison.OrdinalIgnoreCase);
                        p2List[i] = fi.Name.Substring(0, loc);
                        i++;
                    }

                    if (p2List.Count() > 0)
                    {
                        return p2List;
                    }
                    else
                    {
                        return null;
                    }
                }
                catch (Exception e)
                {
                    throw new InvalidProbeException(pcdProbeFolderPath, e);
                }
            }
            else
            {
                throw new MissingProbeFolderException(pcdProbeFolderPath);
            }
        }

   // Various utility methods follow:
        // Validate the existence of a program.
        private bool validateProgram(string prgm)
        {
            if (prgm != "")
            {
                // Check to see if ProgramName exist in application folder, 
                // if not, check in default file path.
                if (System.IO.File.Exists(AppProgFullName(prgm)))
                {
                    return true;
                }
                else
                {
                    // Check to see if ProgramName exist in defaultFilePath.
                    if (System.IO.File.Exists(AppProgShortName(prgm)))
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        // Set up a new PCDMIS part program.
        private void initNewPartProgram(string prgm)
        {
            pcdPartPrograms.CloseAll();
            pcdPartProgram = pcdPartPrograms.Add(prgm, MeasurementUnit,
                pcdDefaultMachine, DefaultProbeName);

            // Remove all but the first eight commands
            pcdCommands = pcdPartProgram.Commands;
            while (pcdCommands.Count > 8)
            {
                pcdCommand = pcdCommands.Item(9);
                pcdCommand.Remove();
            }

            // Set the insertion point at the end of the existing program
            pcdCommand = pcdCommands.Item(pcdCommands.Count);
            pcdCommands.InsertionPointAfter(pcdCommand);

        }

        // This is the path the the program name in the Application folder
        private string AppProgFullName(string prgm)
        {
            return pcdDefaultPcdmisPath + ApplicationFolderName + @"\" + prgm + ".prg";
        }

        // This is the path to the program name in the application folder without the extension
        private string AppProgName(string prgm)
        {
            return pcdDefaultPcdmisPath + ApplicationFolderName + @"\" + prgm;
        }

        // This is the path to the program name in the Default PCDMIS folder
        private string AppProgShortName(string prgm)
        {
            return pcdDefaultPcdmisPath + prgm + ".prg";
        }

        // This is the path to the subprogram name in the subroutine folder
        private string AppProgFullSubrName(string prgm)
        {
            return AppSubrDirName() + @"\" + prgm + ".prg";
        }

        // This is the path to the subprogram name in the subroutine folder without the extension
        private string AppProgSubrName(string prgm)
        {
            return AppSubrDirName() + @"\" + prgm;
        }

        // This is the path to the subroutine folder
        private string AppSubrDirName()
        {
            if (SubroutineFolderName == "")
            {
                return pcdDefaultPcdmisPath + ApplicationFolderName;
            }
            else
            {
                return pcdDefaultPcdmisPath + ApplicationFolderName + @"\" + SubroutineFolderName;
            }
        }

        // Function returns int value as a string.
        private string Fstr(int val)
        {
            return string.Format("{0:F0}", val);
        }

        // Function returns double value as three place decimal string.
        private string Fstr3(double val)
        {
            return string.Format("{0:F3}", val);
        }

        // Function returns double value as five place decimal string.
        private string Fstr5(double val)
        {
            return string.Format("{0:F5}", val);
        }

        // Delete the program from disk.
        private void purgeProgram(string prgm)
        {
            if (prgm != null)
            {
                // Check to see if docFileName exist in application folder and/or 
                // default file path.
                if (System.IO.File.Exists(AppProgFullName(prgm)))
                    System.IO.File.Delete(AppProgFullName(prgm));
                if (System.IO.File.Exists(AppProgShortName(prgm)))
                    System.IO.File.Delete(AppProgShortName(prgm));

                // Now check for existence docFileName in a subroutines folder.
                if (System.IO.Directory.Exists(AppSubrDirName()))
                {
                    if (System.IO.File.Exists(AppProgFullSubrName(prgm)))
                        System.IO.File.Delete(AppProgFullSubrName(prgm));
                }
            }
        }

        /// <summary>
        /// Method to determine if there is a connection to the PCDLRN.
        /// </summary>
        public bool IsSessionActive
        {
            get
            {
                if (pcdSession == null) return false;
                return true;
            }
        }

        /// <summary>
        /// Method that indicates if PCDMIS is available to run a program.
        /// </summary>
        public static Boolean SessionIsAvailable
        { get; set; }

        /// <summary>
        /// Set/Get the application folder name.
        /// </summary>
        public string ApplicationFolderName
        { get; set; }

        /// <summary>
        /// Set/get the folder name which contains subroutines.
        /// </summary>
        public string SubroutineFolderName
        { get; set; }

        /// <summary>
        /// Method to return the default PCDMIS path.
        /// </summary>
        public string DefaultPcdmisPath
        {
            get
            {
                return pcdDefaultPcdmisPath;
            }
        }
   // Methods to insert PCDMIS assignments commands into PCDMIS program file follow:
        public void WriteAssignmentStatement(string varName, string varValue)
        {
            string tmpValue = "\"" + varValue + "\"";
            if (!AddAssignmentCommand(varName, tmpValue)) pcdCommand.Remove();
        }

        public void WriteAssignmentStatement(string varName, int varValue)
        {
            if (!AddAssignmentCommand(varName, Fstr(varValue))) pcdCommand.Remove();
        }

        public void WriteAssignmentStatement(string varName, double varValue)
        {
            if (!AddAssignmentCommand(varName, Fstr5(varValue))) pcdCommand.Remove();
        }

        private Boolean AddAssignmentCommand(string VarName, string VarValue)
        {
            string varName;
            string varValue;
            if (VarName == null | VarName == "")
            {
                varName = "Undefined";
            }
            else
            {
                varName = VarName;
            }
            if (VarValue == null | VarValue == "")
            {
                varValue = "\"Undefined\"";
            }
            else
            {
                varValue = VarValue;
            }

            Boolean returnStatus;
            pcdCommand = pcdCommands.Add(pcdOBT.ASSIGNMENT, true);
            pcdCommand.Marked = true;
            returnStatus = pcdCommand.PutText(varName, pcdEFT.DEST_EXPR, 0);
            returnStatus = pcdCommand.PutText(varValue, pcdEFT.SRC_EXPR, 0);
            return returnStatus;
        }

        public void WriteGenericPoints(string grpName, VectorPoint[] vpnts)
        {
            int pntCnt = vpnts.Count();
            WriteAssignmentStatement(grpName + "_CNT", pntCnt);

            int i = 1;
            foreach (VectorPoint curPnt in vpnts)
            {
                string pntName = grpName + "[" + Fstr(i) + "]";
                writeDmisPoint(pntName, curPnt);
                i++;
            }
        }

        public void WriteGenericPoint(string identifier, VectorPoint vpnt)
        {
            writeDmisPoint(identifier, vpnt);
        }

        private void writeDmisPoint(string pntName, VectorPoint curPnt)
        {
            Boolean retStatus;
            string tmpValue = "ARRAY(" +
                Fstr3(curPnt.X) + "," + Fstr3(curPnt.Y) + "," + Fstr3(curPnt.Z) + "," +
                Fstr5(curPnt.I) + "," + Fstr5(curPnt.J) + "," + Fstr5(curPnt.K) + "," +
                Fstr(curPnt.PntSeqNumber) + ")";

            pcdCommand = pcdCommands.Add(pcdOBT.ASSIGNMENT, true);
            pcdCommand.Marked = true;
            retStatus = pcdCommand.PutText("ARR_" + pntName, pcdEFT.DEST_EXPR, 0);
            retStatus = pcdCommand.PutText(tmpValue, pcdEFT.SRC_EXPR, 0);

        }

   // Methods that control the PCDLRN session follow:
        /// <summary>
        /// Quit the PCDLRN session
        /// </summary>
        public void Quit()
        {
            if (pcdSession != null)
            {
                try
                {
                    pcdSession.Quit();
                    pcdSession = null;
                    Process[] remotePCDLRN = Process.GetProcessesByName("PCDLRN");
                    foreach (Process exe in remotePCDLRN)
                    {
                        exe.Kill();
                    }
                }
                catch (Exception e)
                {
                    string message = e.Message;
                }
            }
        }

        /// <summary>
        /// Set up the connection to the PCDLRN such that when the program that
        /// is connected to PCDLRN quits the PCDLRN automatically quits.
        /// </summary>
        public bool UserExit
        {
            get { if (pcdSession != null) return pcdSession.UserExit; return false; }
            set { if (pcdSession != null) pcdSession.UserExit = value; }
        }

        /// <summary>
        /// Turn on or off Pcdmis Operator mode
        /// </summary>
        public bool OperatorMode
        {
            get { if (pcdSession != null) return pcdSession.OperatorMode; return false; }
            set { if (pcdSession != null) pcdSession.OperatorMode = value; }
        }

        /// <summary>
        /// Make the Pcdmis session visable.
        /// </summary>
        public bool Visable
        {
            get { if (pcdSession != null) return pcdSession.Visible; return false; }
            set { if (pcdSession != null) pcdSession.Visible = value; }
        }

    }

    // Exception classes follow
    public class PcdlrnException : Exception
    {
        public PcdlrnException()
        {
        }

        public PcdlrnException(string message)
            : base(message)
        {
        }

        public PcdlrnException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    public class NoRegistryException : PcdlrnException
    {
        public NoRegistryException()
        {
        }

        public NoRegistryException(string message)
            : base(message)
        {
        }

        public NoRegistryException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    public class NoHaspKeyException : PcdlrnException
    {
        public NoHaspKeyException()
        {
        }

        public NoHaspKeyException(string message)
            : base(message)
        {
        }

        public NoHaspKeyException(string message, Exception inner)
            : base(message, inner)
        {
        }

    }

    public class UnknownPcdlrnErrorException : PcdlrnException
    {
        public UnknownPcdlrnErrorException()
        {
        }

        public UnknownPcdlrnErrorException(string message)
            : base(message)
        {
        }

        public UnknownPcdlrnErrorException(string message, Exception inner)
            : base(message, inner)
        {
        }

    }

    public class InvalidProgramNameException : PcdlrnException
    {
        public InvalidProgramNameException()
        {
        }

        public InvalidProgramNameException(string message)
            : base(message)
        {
        }

        public InvalidProgramNameException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    public class InactivePcdlrnSessionException : PcdlrnException
    {
        public InactivePcdlrnSessionException()
        {
        }

        public InactivePcdlrnSessionException(string message)
            : base(message)
        {
        }

        public InactivePcdlrnSessionException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    public class InvalidProbeException : PcdlrnException
    {
        public InvalidProbeException()
        {
        }

        public InvalidProbeException(string message)
            : base(message)
        {
        }

        public InvalidProbeException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    public class MissingProbeFolderException : PcdlrnException
    {
        public MissingProbeFolderException()
        {
        }

        public MissingProbeFolderException(string message)
            : base(message)
        {
        }

        public MissingProbeFolderException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }
}
